Package: bambu
Type: Package
Title: Context-Aware Transcript Quantification from Long Read RNA-Seq
        data
Version: 3.12.0
Authors@R: c(person("Ying", "Chen", role = c("cre","aut"),
             email = "chen_ying@gis.a-star.edu.sg"),
             person("Andre", "Sim", role = "aut",
             email = "andre_sim@gis.a-star.edu.sg"),
             person("Yuk Kei", "Wan", role = "aut",
             email = "Wan_Yuk_Kei@gis.a-star.edu.sg"),
             person("Jonathan", "Goeke", role = "aut",
             email = "gokej@gis.a-star.edu.sg"))
Description: bambu is a R package for multi-sample transcript discovery and 
    quantification using long read RNA-Seq data. You can use bambu after read 
    alignment to obtain expression estimates for known and novel transcripts and 
    genes. The output from bambu can directly be used for visualisation and 
    downstream analysis such as differential gene expression or transcript
    usage. 
License: GPL-3 + file LICENSE
Encoding: UTF-8
ByteCompile: true
Depends: R(>= 4.1), SummarizedExperiment(>= 1.1.6), S4Vectors(>=
        0.22.1), BSgenome, IRanges
Suggests: AnnotationDbi, Biostrings, rmarkdown, BiocFileCache, ggplot2,
        ComplexHeatmap, circlize, ggbio, gridExtra, knitr, testthat,
        BSgenome.Hsapiens.NCBI.GRCh38,
        TxDb.Hsapiens.UCSC.hg38.knownGene, ExperimentHub (>= 1.15.3),
        DESeq2, NanoporeRNASeq, purrr, apeglm, utils, DEXSeq
Enhances: parallel
SystemRequirements:
biocViews: Alignment, Coverage, DifferentialExpression,
        FeatureExtraction, GeneExpression, GenomeAnnotation,
        GenomeAssembly, ImmunoOncology, LongRead, MultipleComparison,
        Normalization, RNASeq, Regression, Sequencing, Software,
        Transcription, Transcriptomics
bugReports: https://github.com/GoekeLab/bambu/issues
URL: https://github.com/GoekeLab/bambu
RoxygenNote: 7.2.1
LinkingTo: Rcpp, RcppArmadillo
Imports: BiocGenerics, BiocParallel, data.table, dplyr, tidyr,
        GenomeInfoDb, GenomicAlignments, GenomicFeatures,
        GenomicRanges, stats, Rsamtools, methods, Rcpp, xgboost
VignetteBuilder: knitr
git_url: https://git.bioconductor.org/packages/bambu
git_branch: RELEASE_3_22
git_last_commit: 4b0b622
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: yes
Packaged: 2025-10-30 02:48:20 UTC; biocbuild
Author: Ying Chen [cre, aut],
  Andre Sim [aut],
  Yuk Kei Wan [aut],
  Jonathan Goeke [aut]
Maintainer: Ying Chen <chen_ying@gis.a-star.edu.sg>
