# Generated by roxygen2: do not edit by hand

export(bambu)
export(plotBambu)
export(prepareAnnotations)
export(readFromGTF)
export(transcriptToGeneExpression)
export(writeBambuOutput)
export(writeToGTF)
export(trainBambu)
export(compareTranscripts)
importFrom(stats,predict)
importFrom(BiocGenerics,basename)
importFrom(BiocGenerics,unstrand)
importFrom(BiocParallel,bplapply)
importFrom(BiocParallel,bpparam)
importFrom(BiocParallel,bpvec)
importFrom(BiocParallel,bpworkers)
importFrom(BSgenome,getSeq)
importFrom(BSgenome,letterFrequency)
importFrom(BSgenome,available.genomes)
importFrom(BSgenome,getBSgenome)
importFrom(GenomeInfoDb,"seqlevels<-")
importFrom(GenomeInfoDb,keepSeqlevels)
importFrom(GenomeInfoDb,seqlevels)
importFrom(GenomicAlignments,grglist)
importFrom(GenomicAlignments,readGAlignments)
importFrom(GenomicFeatures,exonsBy)
importFrom(GenomicFeatures,transcripts)
importFrom(GenomicRanges,GRanges)
importFrom(GenomicRanges,GRangesList)
importFrom(GenomicRanges,granges)
importFrom(GenomicRanges,intersect)
importFrom(GenomicRanges,makeGRangesListFromDataFrame)
importFrom(GenomicRanges,makeGRangesListFromFeatureFragments)
importFrom(GenomicRanges,match)
importFrom(GenomicRanges,narrow)
importFrom(GenomicRanges,reduce)
importFrom(GenomicRanges,setdiff)
importFrom(GenomicRanges,width)
importFrom(IRanges,shift)
importFrom(Rcpp,sourceCpp)
importFrom(Rsamtools,"yieldSize<-")
importFrom(Rsamtools,BamFile)
importFrom(Rsamtools,BamFileList)
importFrom(Rsamtools,FaFile)
importFrom(Rsamtools,ScanBamParam)
importFrom(Rsamtools,isIncomplete)
importFrom(Rsamtools,indexFa)
importFrom(Rsamtools,scanBamFlag)
importFrom(Rsamtools,yieldSize)
importFrom(SummarizedExperiment,SummarizedExperiment)
importFrom(SummarizedExperiment,cbind)
importFrom(SummarizedExperiment,rbind)
importFrom(xgboost,xgboost)
importFrom(methods,as)
importFrom(methods,is)
importFrom(stats,cor)
importFrom(stats,fisher.test)
importFrom(stats,model.matrix)
importFrom(stats,prcomp)
importFrom(stats,weighted.mean)
importFrom(stats, lm, glm)
importFrom(tidyr,nest)
importFrom(tidyr,pivot_longer)
importFrom(tidyr,pivot_wider)
importFrom(tidyr,separate)
importFrom(tidyr,separate_rows)
importFrom(tidyr,spread)
importFrom(tidyr, replace_na)
importFrom(tidyr, unnest)
import(data.table, except=c(last, first, shift, second, between))
import(dplyr, except=c(last, first, desc, union, setdiff, intersect, slice))
import(IRanges, except=c(slice, collapse, setdiff, intersect,cor))
import(SummarizedExperiment)
import(S4Vectors, except=c(rename, setequal, setdiff, intersect,cor))
useDynLib(bambu, .registration = TRUE)
import(xgboost)
import(BSgenome)
