# Generated by using Rcpp::compileAttributes() -> do not edit by hand
# Generator token: 10BE3573-1514-4C36-9D1C-5A225CD40393

#' EM algorithm
#' @noRd
em_theta <- function(X, Y, maxiter, minvalue, conv) {
    .Call(`_bambu_em_theta`, X, Y, maxiter, minvalue, conv)
}

#' L1-penalized likelihood estimation
#' @noRd
emWithL1 <- function(A, A_full, A_unique, Y, K, maxiter, minvalue, conv) {
    .Call(`_bambu_emWithL1`, A, A_full, A_unique, Y, K, maxiter, minvalue, conv)
}

